==> dovecot-0.9.tar.gz: dovecot-0.9/src/master/login-process.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "iobuffer.h"
#include "fdpass.h"
#include "restrict-access.h"
#include "login-process.h"

#include "../login/master.h"

#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <pwd.h>

typedef struct {
	pid_t pid;
	int fd;
	IO io;
	IOBuffer *outbuf;
} LoginProcess;

static Timeout to;
static HashTable *processes;

static void login_process_destroy(LoginProcess *p);

#define VALIDATE_STR(str) \
	validate_str(str, sizeof(str))

static int validate_str(const char *str, int max_len)
{
	int i;

	for (i = 0; i < max_len; i++) {
		if (str[i] == '\0')
			return TRUE;
	}

	return FALSE;
}

static void login_process_input(void *user_data, int fd, IO io)
{
	const char *env[] = {
		"MAIL", NULL,
		"LOGIN_TAG", NULL,
		NULL
	};
	LoginProcess *p = user_data;
	MasterRequest req;
	MasterReply reply;
	int client_fd, success, ret;

	ret = fd_read(p->fd, &req, sizeof(req), &client_fd);
	if (ret != sizeof(req)) {
		if (ret == 0) {
			/* disconnected, ie. the login process died */
		} else if (ret > 0) {
			/* req wasn't fully read */
			i_error("login: fd_read() couldn't read all req");
		} else {
			i_error("login: fd_read() failed: %s", strerror(errno));
		}

		login_process_destroy(p);
		return;
	}

	/* login process isn't trusted, validate all data to make sure
	   it's not trying to exploit us */
	if (!VALIDATE_STR(req.user) || !VALIDATE_STR(req.pass) ||
	    !VALIDATE_STR(req.chroot_dir) || !VALIDATE_STR(req.mail)) {
		i_error("login: Received corrupted data");
		login_process_destroy(p);
		return;
	}

	if (req.user[0] != '\0' && req.pass[0] != '\0') {
		/* user/pass given, verify that it's valid */

		/*FIXME: success = auth_verify(req.user, req.pass);*/
		req.uid = 1000;
		req.gid = 1000;
		strcpy(req.mail, "maildir:/home/cras/Maildir");
		success = TRUE;
	} else if (set_require_real_user) {
		i_error("login: System user/password not given and "
			"require_real_user setting is on");
		success = FALSE;
	} else {
		success = TRUE;
	}

	if (success) {
		env[1] = req.mail;
		env[3] = req.login_tag;
		success = create_imap_process(client_fd, req.user,
					      req.uid, req.gid,
					      req.chroot_dir, env) != -1;
	}

	/* send reply */
	memset(&reply, 0, sizeof(reply));
	reply.id = req.id;
	reply.success = success;

	if (io_buffer_send(p->outbuf, &reply, sizeof(reply)) < 0)
		login_process_destroy(p);
	else {
		if (success)
			(void)close(client_fd);
	}
}

static LoginProcess *login_process_new(pid_t pid, int fd)
{
	LoginProcess *p;

	PID_ADD_PROCESS_TYPE(pid, PROCESS_TYPE_LOGIN);

	p = i_new(LoginProcess, 1);
	p->pid = pid;
	p->fd = fd;
	p->io = io_add(fd, IO_READ, login_process_input, p);
	p->outbuf = io_buffer_create(fd, default_pool, IO_PRIORITY_DEFAULT,
				     sizeof(MasterReply)*10);

	hash_insert(processes, INT_TO_POINTER(pid), p);
	return p;
}

static void login_process_destroy(LoginProcess *p)
{
	hash_remove(processes, INT_TO_POINTER(p->pid));

	io_buffer_destroy(p->outbuf);
	io_remove(p->io);
	(void)close(p->fd);
	i_free(p);
}

pid_t create_login_process(void)
{
	static char *const argv[] = { NULL };
	struct passwd *pwd;
	pid_t pid;
	int fd[2];

	if ((pwd = getpwnam(set_login_user)) == 0)
		i_fatal("Login user doesn't exist: %s", set_login_user);

	if (pwd->pw_uid == 0)
		i_fatal("Login process must not run as root");

	/* create communication to process with a socket pair */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd) == -1) {
		i_error("socketpair() failed: %s", strerror(errno));
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		(void)close(fd[0]);
		(void)close(fd[1]);
		i_error("fork() failed: %s", strerror(errno));
		return -1;
	}

	if (pid != 0) {
		/* master */
		login_process_new(pid, fd[0]);
		(void)close(fd[1]);
		return pid;
	}

	/* move the listen handle to 0 */
	if (dup2(imap_fd, 0) < 0) {
		i_fatal("login: dup2(%d, 0) failed: %s",
			imap_fd, strerror(errno));
	}
	/* the imap_fd itself is closed by clean_child_process() */

	/* move communication handle to 1 */
	if (dup2(fd[1], 1) < 0) {
		i_fatal("login: dup2(%d, 1) failed: %s",
			fd[1], strerror(errno));
	}
	(void)close(fd[0]);
	(void)close(fd[1]);

	/* don't pass stderr to login */
	(void)close(2);

	clean_child_process();

	/* setup access environment - needs to be done after
	   clean_child_process() since it clears environment */
	restrict_access_set_env(set_login_user, pwd->pw_uid, pwd->pw_gid,
				set_login_chroot);

	execv(set_login_executable, argv);

	/* note that since we've closed syslog, this error message won't show
	   in it unless we're not chroot()ed (/dev/log can't be opened) */
	i_fatal("execv(%s) failed: %s", set_login_executable, strerror(errno));
	return -1;
}

void login_process_died(pid_t pid)
{
	/* ignore, login_process_input() should notice that the process died */
}

static void login_processes_restart(void *user_data, Timeout timeout)
{
	/* create max. one process every second, that way if it keeps
	   dying all the time we don't eat all cpu with fork()ing. */
	if (hash_size(processes) < set_login_processes_count)
                (void)create_login_process();
}

void login_processes_init(void)
{
        processes = hash_create(default_pool, 128, NULL, NULL);
	to = timeout_add(1000, login_processes_restart, NULL);
}

static void login_hash_destroy(void *key, void *value, void *user_data)
{
	login_process_destroy(value);
}

void login_processes_deinit(void)
{
	timeout_remove(to);

	hash_foreach(processes, login_hash_destroy, NULL);
	hash_destroy(processes);
}

==> dovecot-0.92.tar.gz: dovecot-0.92/src/master/login-process.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "iobuffer.h"
#include "fdpass.h"
#include "restrict-access.h"
#include "login-process.h"
#include "auth-process.h"
#include "master-interface.h"

#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>

typedef struct {
	int refcount;

	pid_t pid;
	int fd;
	IO io;
	IOBuffer *outbuf;
	unsigned int destroyed:1;
} LoginProcess;

typedef struct {
	LoginProcess *process;
	int login_id;
	int auth_id;
	int fd;

	char login_tag[LOGIN_TAG_SIZE];
} LoginAuthRequest;

static int auth_id_counter;
static Timeout to;
static HashTable *processes;

static void login_process_destroy(LoginProcess *p);
static void login_process_unref(LoginProcess *p);

static void auth_callback(AuthCookieReplyData *cookie_reply, void *user_data)
{
	const char *env[] = {
		"MAIL", NULL,
		"LOGIN_TAG", NULL,
		NULL
	};
	LoginAuthRequest *request = user_data;
        LoginProcess *process;
	MasterReply reply;
	int success;

	env[1] = cookie_reply->mail;
	env[3] = request->login_tag;

	if (cookie_reply == NULL || !cookie_reply->success)
		success = FALSE;
	else {
		success = create_imap_process(request->fd, cookie_reply->user,
					      cookie_reply->uid,
					      cookie_reply->gid,
					      cookie_reply->home,
					      cookie_reply->chroot, env);
	}

	/* reply to login */
	reply.id = request->login_id;
	reply.success = success;

	process = request->process;
	if (io_buffer_send(process->outbuf, &reply, sizeof(reply)) < 0)
		login_process_destroy(process);

	(void)close(request->fd);
	login_process_unref(process);
	i_free(request);
}

static void login_process_input(void *user_data, int fd, IO io)
{
	LoginProcess *p = user_data;
	AuthProcess *auth_process;
	LoginAuthRequest *authreq;
	MasterRequest req;
	int client_fd, ret;

	ret = fd_read(p->fd, &req, sizeof(req), &client_fd);
	if (ret != sizeof(req)) {
		if (ret == 0) {
			/* disconnected, ie. the login process died */
		} else if (ret > 0) {
			/* req wasn't fully read */
			i_error("login: fd_read() couldn't read all req");
		} else {
			i_error("login: fd_read() failed: %s", strerror(errno));
		}

		login_process_destroy(p);
		return;
	}

	/* login process isn't trusted, validate all data to make sure
	   it's not trying to exploit us */
	if (!VALIDATE_STR(req.login_tag)) {
		i_error("login: Received corrupted data");
		login_process_destroy(p);
		return;
	}

	/* ask the cookie from the auth process */
	authreq = i_new(LoginAuthRequest, 1);
	p->refcount++;
	authreq->process = p;
	authreq->login_id = req.id;
	authreq->auth_id = ++auth_id_counter;
	authreq->fd = client_fd;
	strcpy(authreq->login_tag, req.login_tag);

	auth_process = auth_process_find(req.auth_process);
	if (auth_process == NULL) {
		i_error("login: Authentication process %u doesn't exist",
			req.auth_process);
		auth_callback(NULL, &authreq);
	} else {
		auth_process_request(auth_process, authreq->auth_id, req.cookie,
				     auth_callback, authreq);
	}
}

static LoginProcess *login_process_new(pid_t pid, int fd)
{
	LoginProcess *p;

	PID_ADD_PROCESS_TYPE(pid, PROCESS_TYPE_LOGIN);

	p = i_new(LoginProcess, 1);
	p->refcount = 1;
	p->pid = pid;
	p->fd = fd;
	p->io = io_add(fd, IO_READ, login_process_input, p);
	p->outbuf = io_buffer_create(fd, default_pool, IO_PRIORITY_DEFAULT,
				     sizeof(MasterReply)*10);

	hash_insert(processes, INT_TO_POINTER(pid), p);
	return p;
}

static void login_process_destroy(LoginProcess *p)
{
	if (p->destroyed)
		return;
	p->destroyed = TRUE;

	hash_remove(processes, INT_TO_POINTER(p->pid));
	login_process_unref(p);
}

static void login_process_unref(LoginProcess *p)
{
	if (--p->refcount > 0)
		return;

	io_buffer_destroy(p->outbuf);
	io_remove(p->io);
	(void)close(p->fd);
	i_free(p);
}

static pid_t create_login_process(void)
{
	static const char *argv[] = { NULL, NULL };
	pid_t pid;
	int fd[2];

	if (set_login_uid == 0)
		i_fatal("Login process must not run as root");

	/* create communication to process with a socket pair */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd) == -1) {
		i_error("socketpair() failed: %s", strerror(errno));
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		(void)close(fd[0]);
		(void)close(fd[1]);
		i_error("fork() failed: %s", strerror(errno));
		return -1;
	}

	if (pid != 0) {
		/* master */
		login_process_new(pid, fd[0]);
		(void)close(fd[1]);
		return pid;
	}

	/* move the listen handle to 0 */
	if (dup2(imap_fd, 0) < 0) {
		i_fatal("login: dup2(%d, 0) failed: %s",
			imap_fd, strerror(errno));
	}
	/* the imap_fd itself is closed by clean_child_process() */

	/* move communication handle to 1 */
	if (dup2(fd[1], 1) < 0) {
		i_fatal("login: dup2(%d, 1) failed: %s",
			fd[1], strerror(errno));
	}
	(void)close(fd[0]);
	(void)close(fd[1]);

	/* don't pass stderr to login */
	(void)close(2);

	clean_child_process();

	/* setup access environment - needs to be done after
	   clean_child_process() since it clears environment */
	restrict_access_set_env(set_login_user, set_login_uid, set_login_gid,
				set_login_chroot);

	argv[0] = set_login_executable;
	execv(argv[0], (char **) argv);

	i_fatal("execv(%s) failed: %s", argv[0], strerror(errno));
	return -1;
}

static void login_processes_start_missing(void *user_data, Timeout timeout)
{
	/* create max. one process every second, that way if it keeps
	   dying all the time we don't eat all cpu with fork()ing. */
	if (hash_size(processes) < set_login_processes_count)
                (void)create_login_process();
}

void login_processes_init(void)
{
        auth_id_counter = 0;
        processes = hash_create(default_pool, 128, NULL, NULL);
	to = timeout_add(1000, login_processes_start_missing, NULL);
}

static void login_hash_destroy(void *key, void *value, void *user_data)
{
	login_process_destroy(value);
}

void login_processes_deinit(void)
{
	timeout_remove(to);

	hash_foreach(processes, login_hash_destroy, NULL);
	hash_destroy(processes);
}
